#!/usr/bin/env node
'use strict';

function readConfigAndRun(args) {
  var fs = require('fs');
  var path = require('path');
  var cwd = args.cwd || process.cwd();
  var text;
  var filename;
  var config;

  if (args.config) {
    text = fs.readFileSync(path.join(cwd, args.config), 'utf8');
  }
  else {
    filename = path.join(cwd, 'Goldilocks.yml');

    if (fs.existsSync(filename)) {
      text = fs.readFileSync(filename, 'utf8');
    }
    else {
      filename = path.join(cwd, 'Goldilocks.json');
      if (fs.existsSync(filename)) {
        text = fs.readFileSync(filename, 'utf8');
      } else {
        text = '{}';
      }
    }
  }

  try {
    config = JSON.parse(text);
  } catch(e) {
    try {
      config = require('js-yaml').safeLoad(text);
    } catch(e) {
      throw new Error(
        "Could not load '" + filename + "' as JSON nor YAML"
      );
    }
  }

  require('../lib/goldilocks.js').create(config);
}

if (process.argv.length === 2) {
  readConfigAndRun({});
}
else if (process.argv.length === 4) {
  if ('-c' === process.argv[3] || '--config' === process.argv[3]) {
    readConfigAndRun({ config: process.argv[4] });
  }
}
else if (process.argv.length > 2) {
  var program = require('commander');

  program
    .version(require('package.json').version)
    .option('--config', 'Path to config file (Goldilocks.json or Goldilocks.yml) example: --config /etc/goldilocks/Goldilocks.json')
    .option('--tunnel [token]', 'Turn tunnel on. This will enter interactive mode for login if no token is specified.')
    .parse(process.argv);

  readConfigAndRun(program);
}
else {
  throw new Error("impossible number of arguments: " + process.argv.length);
}
